/*	RTT measurement 
	Coded By: Rohit Srivastava (rohit@comports.com)
		  http://members.tripod.com/~srohit
	On: I-don't-remember
*/

#ifndef __INCLUDE_RTT__
#include <rtt.h>
#endif

#include <sys/param.h>

#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/socket.h>
#include <sys/times.h>

#include <sys/stat.h>

#include <fcntl.h>
#include <setjmp.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>

#include <hostoaddr.h>
#include <checksum.h>

#include <iostream.h>

static int sock ;
static struct sockaddr * daddr;
static long *rtt = 0 ;
int rtt_size = 0;

static float crtt , mrtt  ;
static float srtt = T_OUT ;


static sigjmp_buf env ;

unsigned long itime;

void RttSighand(int);
void RttStop(int);

RttAnalyser::RttAnalyser()
{
	timeout = T_OUT;
	nu_paks = NU_PAKS;
	stime = T_BURST;

	flag = 0;

	destination = 0;

	seq = 1;
	id = (0xffff & (int)getpid() );
	
	rtt = new long[nu_paks];
	rtt_size = nu_paks;

	return;
}

void RttAnalyser::set_destination(char *dest)
{
	destination = new char[strlen(dest) + 1];
	strcpy(destination , dest);
	return;
}

void RttAnalyser::set_nupaks(int n)
{
	nu_paks = n;
	
	if( rtt != 0 )
	{
		delete(rtt);
		rtt = new long[nu_paks];
		rtt_size = nu_paks ;
	}
	return;
}

/* stime = sleep time */

void RttAnalyser::set_stime(int n)
{
	stime = n;
	return;
}

void RttAnalyser::set_timeout(int n)
{
	timeout = n;
	return;
}

void RttAnalyser::set_tries(int n)
{
	nu_tries = n;
	flag |= NTRIES_SET;
	return;
}

int RttAnalyser::get_rtt( float *lmrtt , float * lsrtt )
{
	
	flag |= ONLY_RTT ;
	
	if( !(flag & NTRIES_SET) )
	{
		cout << "Nu of tries not set" << endl;
		exit(1);
	}
	
	// Nu tries set now call srt_graph
	
	srt_graph();
	
	*lmrtt = mrtt;
	*lsrtt = srtt ;

	// Now return
	return 0;
}
	

int RttAnalyser::srt_graph()
{
	int i,
	    size = sizeof(struct icmphdr) + sizeof(long) ,
	    fd ;
	

	unsigned long addr = 0;

	char on = 1,
	     *buf = new char[size];

	struct sigaction sg;

	// Host to internal IP-Addr form

	if( hostToaddr( destination , &addr ) < 0 )
	{
	cout << "Host to addr conversion failed!" << endl;
	(void)exit(1);
	}

	struct sockaddr_in taddr;
	bzero( (void *)&taddr , sizeof(taddr));
	taddr.sin_addr.s_addr = addr;
	taddr.sin_family = AF_INET;
	taddr.sin_port = 0;
	
	daddr = (struct sockaddr *)&taddr;
		
	// make a fifo 
	/*if( mkfifo( FILE_PATH , (short)0644 ) < 0 )
	{
		perror("Unable to make fifo!");
		(void)exit(1);
	}
	*/
	// Now it will send all the data to this fifo 
	


	if( (fd = open(FILE_PATH , O_WRONLY | O_CREAT ,(short)0600 )) < 0 )
	{
		perror("What the hell! Error in open.");
		(void)exit(1);
	}
		

	// Now get socket 
	if( (sock = socket( AF_INET , SOCK_RAW , IPPROTO_ICMP)) < 0 )
	{
		perror("Unable to allocate raw socket!");
		(void)exit(1);
	}
	 
	// Now set the signal handler

	sigemptyset(&sg.sa_mask);
	sg.sa_flags = 0;
	sg.sa_handler = RttSighand;
	sg.sa_restorer = 0;

	if( sigaction(SIGALRM , &sg , 0) < 0 )
	{
		perror("Error in sigaction");
		(void)exit(1);
	}

	sg.sa_handler = RttStop;
		
	if( sigaction( SIGINT , &sg , 0) < 0 )
	{
		perror("Error in sigaction");
		(void)exit(1);
	}
	if( sigaction( SIGQUIT , &sg ,0) < 0 )
	{
		perror("Error in sigaction");
		(void)exit(1);
	}


	// Now make the paket
	bzero( buf , size );
	struct icmphdr * icmpHdr = (struct icmphdr *)buf;


	// Initialize the fields
	icmpHdr->type = ICMP_ECHO;
	icmpHdr->code = 0;
	icmpHdr->un.echo.id = id;
	icmpHdr->un.echo.sequence = seq;
	icmpHdr->checksum = 0;
	
	// Compute the checksum
	icmpHdr->checksum = in_cksum((unsigned short *)buf , size);

	if( flag & ONLY_RTT )
	{
	i = nu_tries;
	}
	else
	{
	i = 0xffffff;
	}

	for(i = nu_tries; i >= 0 ; i-- )
	{
		if( !(flag & ONLY_RTT) )
		{
		if( i == 0 )
		i = 0xffffff;
		}
		
		if(sigsetjmp(env,1) == 0)
		{	
		alarm(timeout);
		
		// Send the burst
		sendpaks(buf , size);

		// Recv the burst
		recvpaks();

		// Now reset the alarm
		alarm(0);

		// Sleep now
		sleep(stime);
	
		cout << "Rtt -> " << crtt << " msec" << endl;
	
		// Write the rtt
		if( !(flag & ONLY_RTT) )
		write(fd , (char *)&crtt , sizeof(crtt)); // Errors are ignored

		// Remake the packet
		seq++;
		icmpHdr->type = ICMP_ECHO;
		icmpHdr->code = 0;
		icmpHdr->un.echo.sequence = seq;
		icmpHdr->un.echo.id = id;
		icmpHdr->checksum = 0;

		// Recompute the checksum
		icmpHdr->checksum = in_cksum((unsigned short *)buf , size);

		}
		else
		{
		seq++;
		icmpHdr->type = ICMP_ECHO;
		icmpHdr->code = 0;
		icmpHdr->un.echo.sequence = seq;
		icmpHdr->checksum = 0;

		icmpHdr->checksum = in_cksum((unsigned short *)buf , size);

		if( !(flag & ONLY_RTT) )
		write(fd , (char *)&crtt , sizeof(crtt));

		cout << "Rtt -> " << crtt << endl ;
		}	
	}

	return 0;
}

int RttAnalyser::sendpaks(const char * buf , int size)
{
	struct tms ctm;

	struct icmphdr *icmpHdr = (struct icmphdr *)buf;
	
	itime = times(&ctm);

	// All packets in one burst will have same time

	for( int i = nu_paks ; i > 0 ; i-- )
	{
	if( sendto(sock , buf , size , 0 , daddr , sizeof(struct sockaddr)) < 0 )
		{
			perror("error in sendto");
			(void)exit(1);
		}
		
	};
}

int RttAnalyser::recvpaks()
{
	struct sockaddr addr;
	struct tms ctm ;
	
	int i ;

	int size = sizeof(struct icmphdr) + sizeof(struct iphdr);
	int len ;

	char *buf = new char[size];
	bzero( buf , size );

	struct icmphdr *icmpHdr = (struct icmphdr *)(buf + sizeof(struct iphdr));

	bzero( (void *)rtt , rtt_size*sizeof(long) );

	for(i = nu_paks; i > 0 ; i--)
	{
	if( recvfrom(sock, buf , size , 0 , 0 , &len) < 0 )
	{
		perror("error in recvfrom");
		(void)exit(1);
	}
	
	
	if( icmpHdr->un.echo.sequence != seq || icmpHdr->un.echo.id != id)
	{
#ifdef DEBUG
	cout << "duplicate recv'd" << endl ;
#endif
	continue;     
	}
	

	
	// Now subtract it with current time and store

	*(rtt + i - 1) = (unsigned long)(times(&ctm)) - itime;
	}

	// Now calculate the average and others 

	for( i = 0 ; i < rtt_size ; i++ )
	{
	crtt += *(rtt + i);
	}
	crtt = crtt/(float)rtt_size;

	mrtt = (mrtt + crtt)/2.0;

	srtt = srtt*Alpha + (1 - Alpha)*(crtt);


	// Return now
	return 0;
}

RttAnalyser::~RttAnalyser()
{
	delete(destination);
	return;
}


// Timeout signal handler 
void RttSighand(int iar)
{
	int i ;

	// It reached here..anyway..

	// Now to check how many are recv'd 

	int denom = rtt_size;

	for(i = 0 ; i < rtt_size ; i++ )
	{
		if( *(rtt + i) == 0 ) denom--;
	}

	if( denom == 0 ) // Hmmm no packets recv'd.. total loss
	{
	crtt = -1;
	siglongjmp(env,1);
	}

	for(i = 0; i < rtt_size ; i++ )
	{
	crtt += *(rtt + i);
	}
	crtt = crtt/(float)denom;

	// Currenty leave mrtt and srtt unchanged....

	siglongjmp(env,1);
}	
		
	
void RttStop(int iar)
{
	cout << "\n\n\n-----RttAnalyser statistics------\n";
	cout << "Average Rtt = " << mrtt << " msec" << endl;
	cout << "Smoothed Rtt = " << srtt << " msec" << endl;
	cout << "Rto = " << 2*srtt << " msec" << endl;

	unlink(FILE_PATH);	
	(void)exit(0);
}
